#!/bin/env python3

import argparse
import json
import math
import os
import sys
import time
import shutil

import cv2
import numpy as np
from scipy.spatial.transform import Rotation as R

NUM_PIC = 100
IMG_DIR_NAME = 'images'


def variance_of_laplacian(image):
    return cv2.Laplacian(image, cv2.CV_64F).var()


def sharpness(imagePath):
    image = cv2.imread(imagePath)
    gray = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    fm = variance_of_laplacian(gray)
    return fm


def get_tf(tokens, postprocess):
    # format: ts, fx, fy, m11, m21, m31, m41, m12, ...
    tf = list(map(float, tokens[3:]))
    m = [
        tf[0:4],
        tf[4:8],
        tf[8:12],
        tf[12:16]
    ]

    if not postprocess:
        return m
    else:
        # some post-processing trick
        c2w = np.linalg.inv(m)
        flip_mat = np.array([
            [1, 0, 0, 0],
            [0, -1, 0, 0],
            [0, 0, -1, 0],
            [0, 0, 0, 1]
        ])
        c2w = np.matmul(c2w, flip_mat)  # flip cameras (it just works)
        c2w[:3, 3] *= -1 # negate translate vector

        return c2w.tolist()


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--input', required=True)
    parser.add_argument('--output', required=True)
    parser.add_argument('--postprocess', action='store_true')

    args = parser.parse_args()

    record_dir = args.input
    if not os.path.exists(record_dir) or not os.path.isdir(record_dir):
        sys.exit(f'directory {record_dir} does not exit!')

    img_dir = os.path.join(record_dir, 'PV')
    record_ts = os.path.basename(args.input)

    print(f'start processing {record_dir}, postprocess={args.postprocess}')
    start = time.time()

    output_dir = os.path.join(args.output, record_ts)

    # create output dir
    # if os.path.exists(output_dir) and os.path.isdir(output_dir):
    #     print(f"delet existing folder {output_dir}")
    #     shutil.rmtree(output_dir)

    # output_img_dir = os.path.join(output_dir, IMG_DIR_NAME)
    # os.makedirs(output_dir)
    # os.makedirs(output_img_dir)

    # transforms.json
    frames = []

    with open(os.path.join(args.input, record_ts + '_pv.txt'), 'r') as f:
        lines = f.readlines()
        del lines[1:60]
        num_imgs = len(lines) - 1
        chunk = math.ceil(num_imgs / NUM_PIC)

        # calculate camera intrinsics
        camera_info = lines[0].split(',')
        fl_x, fl_y, frame_w, frame_h = map(float, camera_info)  # unpack
        angle_x = math.atan(frame_w / (fl_x * 2)) * 2
        angle_y = math.atan(frame_h / (fl_y * 2)) * 2
        fovx = angle_x * 180 / math.pi
        fovy = angle_y * 180 / math.pi
        print(
            f"camera:\n\tres={frame_w,frame_h}\n\tfocal={fl_x,fl_y}\n\tfov={fovx,fovy}")
        print(f'total {num_imgs} images, sample every {chunk}')

        num_chosen_imgs = 0
        for i in range(1, len(lines), chunk):
            # pick the image with highest sharpness
            max_sharpness = 0
            chosen_img = None
            chosen_tf = None

            for line in lines[i:i + chunk]:
                tokens = line.split(',')
                img_name = tokens[0] + '.png'
                img_sharpness = sharpness(os.path.join(img_dir, img_name))
                if img_sharpness > max_sharpness:
                    max_sharpness = img_sharpness
                    chosen_img = img_name
                    chosen_tf = get_tf(tokens, args.postprocess)

            print(f'choose image {chosen_img}, sharpness: {max_sharpness}')

            # shutil.copyfile(
            #     os.path.join(img_dir, chosen_img),
            #     os.path.join(output_img_dir, chosen_img))
            
            num_chosen_imgs += 1

            # calculate quaternion and translation
            tf_mtx = np.array(chosen_tf)
            qvec = R.from_matrix(tf_mtx[:3,:3]).as_quat()
            qvec = qvec[..., [3,0,1,2]] # change convention from xyzw to wxyz
            tvec = tf_mtx[:3,3]
            frames.append(
                [num_chosen_imgs, qvec, tvec, chosen_img]
            )

    # dump transformation
    with open(os.path.join(args.output, 'images.txt'), 'w') as f:
        for frame in frames:
            img_id, qvec, tvec, img_name = frame
            f.write(f'{img_id} ')
            f.write(f'{" ".join(qvec.astype(str).tolist())} ')
            f.write(f'{" ".join(tvec.astype(str).tolist())} ')
            f.write(f'1 ') # camera id
            f.write(f'{img_name}\n\n')

    end = time.time()
    print(
        f'finish processing {record_dir}, {num_chosen_imgs} chosen, duration: {(end - start):.2f} s')
